#include <fstream>
#include <cstdio>
#include <vector>
#include <string>
#include <argh.hh>

using namespace std;

#define BGR24 0x42475218  // BGR24 fourcc
#define BGR16 0x42475210  // BGR16 fourcc
#define BGR15 0x4247520F  // BGR15 fourcc

static unsigned FOURCC;

static FILE *fout;
static unsigned frameus = 1000000 / 25;
static vector<string> files;

static struct audio
{
    unsigned rate, chancount, bitness;
    double spacing;
    string file;
    
    unsigned bytecount(double seconds) const
    {
        unsigned samplecount = (unsigned)(seconds * rate);
        return samplecount * chancount * (bitness/8);
    }
    double time(unsigned bytes) const
    {
        bytes /= chancount;
        bytes /= (bitness/8);
        return bytes / (double)rate;
    }
    
    FILE *fp;
    void open()
    {
        fp = NULL;
        if(file == "-")
            fp = stdin;
        else if(file.size())
        {
            fp = fopen(file.c_str(), "rb");
            if(!fp)
            {
                perror(file.c_str());
                file = "";
            }
        }
        if(fp && ferror(fp))
            perror(file.c_str());
    }
} audio = {44100,2,16, 0, "", NULL};

static unsigned width, height, framecount, bytesPerPixel;

static bool xflip = false;
static bool yflip = false;

static void dwf(unsigned dw) /* Write dword */
{
    fputc((unsigned char)((dw      ) & 255), fout);
    fputc((unsigned char)((dw >> 8 ) & 255), fout);
    fputc((unsigned char)((dw >> 16) & 255), fout);
    fputc((unsigned char)((dw >> 24) & 255), fout);
}
static void wf(unsigned short dw) /* Write word */
{
    fputc((unsigned char)((dw      ) & 255), fout);
    fputc((unsigned char)((dw >> 8 ) & 255), fout);
}

class SizeHere
{
    unsigned pos;
public:
    SizeHere()
    {
        pos = ftell(fout);
        fwrite("----", 4, 1, fout);
    }
    ~SizeHere()
    {
        unsigned now = ftell(fout);
        fseek(fout, pos, SEEK_SET);
        dwf(now - pos - 4);
        fseek(fout, now, SEEK_SET);
    }
};

class AVIupdater
{
    char taskcode;
    unsigned bytes;
    long taskpos;
public:
    AVIupdater(char code, unsigned size) : taskcode(code), bytes(size)
    {
        taskpos = ftell(fout);
        if(size==2) wf(0);
        else if(size==4) dwf(0);
    }
    ~AVIupdater()
    {
        long waspos = ftell(fout);
        unsigned value = 0;
        fseek(fout, taskpos, SEEK_SET);
        
        switch(taskcode)
        {
            case 1: value = width; break;
            case 2: value = height; break;
            case 3: value = width * height * bytesPerPixel; break;
            case 4: value = framecount; break;
            case 5: value = (unsigned) (width*height * 3e6 / frameus); break;
            case 6: value = FOURCC; break;
            default:
            {
                fprintf(stderr, "AVIupdater: taskcode=%d - uhh?\n", taskcode);
                throw;
            }
        }
        
        if(bytes==2) wf(value);
        else if(bytes==4)dwf(value);
        
        fseek(fout, waspos, SEEK_SET);
    }
};

static vector<AVIupdater> tasks;

static bool isValidTGA(FILE* fp)
{
    // Check that it's an uncompressed 24-bit tga:
    fseek(fp, 2, SEEK_SET);
    int tgaType = fgetc(fp);
    fseek(fp, 16, SEEK_SET);
    int bitsPerPixel = fgetc(fp);
    if(tgaType != 2 || bitsPerPixel != 24) return false;
    return true;
}

static void CalcSize()
{
    audio.open();

    width = 0;
    height = 0;
    
    for(unsigned i=0; i<files.size(); ++i)
    {
        if(files[i] == "-")
        {
            /*setbuf(stdin, NULL);*/
            continue;
        }
        
        FILE *fp = fopen(files[i].c_str(), "rb");
        if(!fp)
        {
            perror(files[i].c_str());
            continue;
        }
        if(isValidTGA(fp))
        {
            fseek(fp, 12, SEEK_SET);
            width = fgetc(fp);  width |= fgetc(fp) << 8;
            height = fgetc(fp); height |= fgetc(fp) << 8;
            fclose(fp);
            return;
        }
        fclose(fp);
    }
}

static void WriteFile(const char *riffhdr, const string &fn)
{
    FILE *fp;
    
    if(fn == "-")
        fp = stdin;
    else
    {
        fp = fopen(fn.c_str(), "rb");
        if(!fp)
        {
            perror(fn.c_str());
            return;
        }
    }

    int datapos = fgetc(fp);
    if(fn == "-" && datapos == EOF)
    {
        // End of files in stdin
        return;
    }
    
    datapos += 18;
    
    unsigned char *HdrBuf = (unsigned char *)alloca(datapos);
    
    fread(HdrBuf+1, 1, datapos-1, fp);
    
    int tgaType = HdrBuf[2];
    unsigned new_width  = HdrBuf[12];  new_width |= HdrBuf[13] << 8;
    unsigned new_height = HdrBuf[14]; new_height |= HdrBuf[15] << 8;

    int bitsPerPixel = HdrBuf[16];
    
    if(tgaType != 2)
    {
        fprintf(stderr, "%s is not an uncompressed tga.\n", fn.c_str());
        if(fn != "-")fclose(fp);
        return;
    }
    
    switch(bitsPerPixel)
    {
        case 15: FOURCC = BGR15; break;
        case 16: FOURCC = BGR16; break;
        case 24: FOURCC = BGR24; break;
        default:
        {
            fprintf(stderr, "%s is not 15/16/24 -bit tga.\n", fn.c_str());
            if(fn != "-")fclose(fp);
            return;
        }
    }
    
    bytesPerPixel = (bitsPerPixel+7)/8;

    if(!width && !height)
    {
        width = new_width;
        height = new_height;
        fprintf(stderr, "Geometry is %ux%u\n", width, height);
    }
    if(new_width != width || new_height != height)
    {
        fprintf(stderr, "%s (%ux%u) had different size than the first "
                "valid image (%ux%u)\n", fn.c_str(),
                new_width, new_height, width, height);
        if(fn == "-")
        {
            char Buf[4096];
            unsigned left = new_width * new_height;
            while(left > 0)
            {
                unsigned now = left;
                if(now > sizeof Buf)now = sizeof Buf;
                fread(Buf, 1, now, stdin);
                left -= now;
            }
        }
        else
            fclose(fp);
        return;
    }
    
    fwrite(riffhdr, 4, 1, fout);
    SizeHere tmp;
    
    if(yflip && fn == "-")
    {
        char *Buf = new char[width*height*bytesPerPixel];
        for(unsigned y=0; y<height; ++y)
        {
            unsigned yy = height - y - 1;
            fread(&Buf[yy*width*bytesPerPixel], bytesPerPixel, width, fp);
        }
        fwrite(Buf, bytesPerPixel, height*width, fout);
        delete[] Buf;
    }
    else
    {
        char *Buf = new char[width*bytesPerPixel];
        for(unsigned y=0; y<height; ++y)
        {
            unsigned yy = yflip ? (height - y - 1) : y;
            
            if(fn != "-")
                fseek(fp, datapos + yy*width*bytesPerPixel, SEEK_SET);
            fread(Buf, bytesPerPixel, width, fp);
            
            if(xflip)
                for(unsigned x=width; x-->0; )fwrite(&Buf[x*bytesPerPixel], 1, bytesPerPixel, fout);
            else
                fwrite(Buf, bytesPerPixel, width, fout);
        }
        delete[] Buf;
    }
    if(fn != "-")
        fclose(fp);
}

static void WriteSTRHvids()
{
    framecount = 0;
    
    fwrite("strh", 4, 1, fout);
    SizeHere tmp;
    fwrite("vids", 4, 1, fout);
    tasks.push_back(AVIupdater(6, 4)); //FOURCC
    dwf(0);     // flags
    wf(0);      // priority
    wf(0);      // language
    dwf(0);     // initial frames
    dwf(1000000);                  // scale
    dwf((unsigned)(1e12/frameus)); // rate  (put scale*fps here)
    dwf(0);     // start
    dwf(0);     // scale*framecount
    dwf(1048576);// suggested buffer size
    dwf(10000); // quality
    tasks.push_back(AVIupdater(3, 4));
    wf(0);       // left
    wf(0);       // top
    tasks.push_back(AVIupdater(1, 2)); //right
    tasks.push_back(AVIupdater(2, 2)); //bottom
}
static void WriteSTRFvids()
{
    fwrite("strf", 4, 1, fout);
    SizeHere tmp;
    dwf(40);    // size - ????
    tasks.push_back(AVIupdater(1, 4)); //width
    tasks.push_back(AVIupdater(2, 4)); //height
    wf(1);      // planes
    wf(24);     // depth
    /* compression type */
    tasks.push_back(AVIupdater(6, 4)); //FOURCC
    tasks.push_back(AVIupdater(3, 4)); // width*height*3
    dwf(0);     // ?
    dwf(0);     // ?
    dwf(0);     // ?
    dwf(0);     // ?
}
static void WriteSTRHauds()
{
    fwrite("strh", 4, 1, fout);
    SizeHere tmp;
    fwrite("auds", 4, 1, fout);
    
    dwf(0);     // FOURCC for audio
    dwf(0);     // Handler
    dwf(0);     // flags
    wf(0);      // priority
    wf(0);      // language
    dwf(0);     // initial frames
    dwf(1);     // scale
    dwf(audio.rate); // rate
    dwf(0);     // start
    dwf(0);     // length (audio.rate * kesto)
    dwf(1048576);// suggested buffer size
    dwf(0);     // quality
    dwf(audio.chancount * (audio.bitness/8)); // sample size
    wf(0); wf(0); wf(0); wf(0); // frame size
}
static void WriteSTRFauds()
{
    fwrite("strf", 4, 1, fout);
    SizeHere tmp;
    wf(1);               // PCM format
    wf(audio.chancount); // Channel count
    dwf(audio.rate);     // Rate
    dwf(audio.bytecount(1)); // Samples per second
    wf(audio.chancount * (audio.bitness/8)); // Block align
    wf(audio.bitness);   // Bitness
    wf(0);               // cbSize
}
static void WriteAVIH()
{
    unsigned streamcount = 1;
    if(audio.file.size())
        ++streamcount;

    fwrite("avih", 4, 1, fout);
    SizeHere tmp;
    dwf(frameus);   // microseconds per frame
    tasks.push_back(AVIupdater(5, 4)); //max bytes per sec
    dwf(0);         // padding (0)
    dwf(0);         // flags
    tasks.push_back(AVIupdater(4, 4)); //framecount
    dwf(0);           // initial frames
    dwf(streamcount); // streamcount
    dwf(1048576);     // suggested buffer size
    tasks.push_back(AVIupdater(1, 4)); //width
    tasks.push_back(AVIupdater(2, 4)); //height
    dwf(0);dwf(0);dwf(0);dwf(0); // reserved
}
static void WriteSTRLvids() /* Stream list */
{
    /* STRL is a "strl" LIST */
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("strl", 4, 1, fout);
    WriteSTRHvids();
    WriteSTRFvids();
}
static void WriteSTRLauds() /* Stream list */
{
    /* STRL is a "strl" LIST */
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("strl", 4, 1, fout);
    WriteSTRHauds();
    WriteSTRFauds();
}
static void WriteHDRL() /* Header list */
{
    /* HDRL is a "hdrl" LIST */
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("hdrl", 4, 1, fout);
    WriteAVIH();
    WriteSTRLvids();
    if(audio.file.size())
    {
        WriteSTRLauds();
    }
}
static void WriteMOVI() /* Write chunks */
{
    /* MOVI is a "movi" LIST */
    fwrite("LIST", 4, 1, fout);
    SizeHere tmp;
    fwrite("movi", 4, 1, fout);
    
    double videotime = audio.spacing;
    for(unsigned a=0; a<files.size(); ++a)
    {
        if(audio.fp && videotime >= audio.spacing)
        {
            unsigned bytes = audio.bytecount(videotime);
            
            char *Buf = new char[bytes];
            
            int k = fread(Buf, 1, bytes, audio.fp);
            if(ferror(audio.fp))
            {
                perror(audio.file.c_str());
                fclose(audio.fp);
                audio.fp = NULL;
            }
            else if(feof(audio.fp))
            {
                fprintf(stderr, "Error: audio ended before images did.\n");
                fclose(audio.fp);
                audio.fp = NULL;
            }
            
            fwrite("01wb", 1, 4, fout); // 01 = stream index
            SizeHere tmp;
            fwrite(Buf, 1, k, fout);
            
            delete[] Buf;
            
            videotime -= audio.time(k);
        }
        if(1)
        {
            if(files[a] == "-")
            {
                if(feof(stdin))
                    continue;
                /*
                int c = getc(stdin);
                if(c == EOF)continue;
                ungetc(c, stdin);
                */
            }
            WriteFile("00dc", files[a]); // 00 = stream index
            if(files[a] == "-")
            {
                --a;
            }
        }
        
        videotime += frameus / 1E6;
        ++framecount;
    }
}

static void WriteAVI()
{
    fwrite("RIFF", 4, 1, fout);
    SizeHere tmp;
    fwrite("AVI ", 4, 1, fout);
    WriteHDRL();
    WriteMOVI();
    tasks.clear();
    fprintf(stderr, "Wrote %u video frame%s\n", framecount, framecount==1?"":"s");
}

bool readInputFilenames(const string& listfile)
{
    ifstream is(listfile.c_str());
    if(!is)
    {
        fprintf(stderr, "Can't open "); perror(listfile.c_str());
        return false;
    }

    string filename;

    while(is)
    {
        getline(is, filename);
        if(is.good() && filename.size() && filename[0]!='#')
        {
            while(filename.size() && filename[0]==' ') filename.erase(0,1);
            while(filename.size() && filename[filename.size()-1]==' ')
                filename.erase(filename.size()-1,1);
            if(filename.size())
                files.push_back(filename);
        }
    }

    return true;
}

int main(int argc, const char *const *argv)
{
    string outfile;
    
    ParamHandler Argh;
    Argh.AddLong("help",    'h').SetBool().SetDesc("This help.");
    Argh.AddLong("version", 'V').SetBool().SetDesc("Displays version information.");
    Argh.AddLong("outfile", 'o').SetString(1).SetDesc("Specify the output file.", "<filename>");
    Argh.AddLong("fps",     'f').SetFloat(23e-5, 1e6).SetDesc("Frame audio.rate (default: 25).", "<float>");
    Argh.AddLong("xflip",   'X').SetBool().SetDesc("Flip horizontally (not default)");
    Argh.AddLong("yflip",   'Y').SetBool().SetDesc("Flip vertically (default)");
    Argh.AddLong("list",    '@').SetString(1).SetDesc("Read input file names from the given file.", "<filename>");
    Argh.AddLong("audio",   'a').SetString(1).SetDesc("Specify raw audio source (can be a pipe).", "<filename>");
    Argh.AddLong("mono",    'n').SetBool().SetDesc("Audio is mono (not stereo).");
    Argh.AddLong("byte",    'b').SetBool().SetDesc("Audio is 8-bit (not 16-bit).");
    Argh.AddLong("rate",    'r').SetInt(18,999999).SetDesc("Specify audio rate in hertz.", "<rate>");
    Argh.AddLong("audiochunksize",
                            's').SetFloat(1e-3,60).SetDesc("Specify audio spacing.", "<seconds>");
    
    Argh.StartParse(argc, argv);
    for(;;)
    {
        int c = Argh.GetParam();
        if(c == -1)break;
        switch(c)
        {
            case 'V': printf("%s\n", VERSION); return 0;
            case 'h':
                printf(
                    "This is animmaker-"VERSION"\n"
                    "Copyright (C) 1992,2002 Bisqwit (http://bisqwit.iki.fi/)\n"
                    "\nUsage: animmaker [<option> [<...>]] <file> [<...>]\n"
                    "\nThis software assumes the files are uncompressed rgb24 tga files.\n"
                    "\nOptions:\n");
                Argh.ListOptions();
                printf(
                    "\n"
                    "You can also give the tga files in stdin, example:\n"
                    " cat *.tga | animmaker -ogruu.avi -\n"
                    "\n"
                    "No warranty whatsoever.\n");
                return 0;
            case 'o':
                outfile = Argh.GetString();
                break;
            case 'X':
                xflip = Argh.GetBool();
                break;
            case 'Y':
                yflip = Argh.GetBool();
                break;
            case 'f':
                frameus = (unsigned)(1e6 / Argh.GetFloat());
                break;
            case '@':
                if(!readInputFilenames(Argh.GetString())) return -1;
                break;
            case 'a':
                if(audio.file.size())
                {
                    fprintf(stderr, "Error: Can not specify more than one audio source!\n");
                    return -1;
                }
                audio.file = Argh.GetString();
                break;
            case 'r':
                audio.rate = Argh.GetInt();
                break;
            case 'm':
                audio.chancount == Argh.GetBool() ? 1 : 2;
                break;
            case 'b':
                audio.bitness = Argh.GetBool() ? 8 : 16;
                break;
            case 's':
                audio.spacing = Argh.GetFloat();
                break;
            default:
                files.push_back(Argh.GetString());
        }
    }
    if(!Argh.ok())return -1;
    if(!files.size())
    {
        fprintf(stderr, "Error: At least one file must be specified. Try --help.\n");
        return -1;
    }
    
    if(audio.file == "-")
        for(unsigned f=0; f<files.size(); ++f)
            if(files[f] == "-")
            {
                fprintf(stderr, "Error: You can not read both audio and video from the same stdin. Sorry :(\n");
                return -1;
            }
    
    if(!outfile.size() || outfile == "-")
    {
        fout = stdout;
    }
    else
        fout = fopen(outfile.c_str(), "wb");
    if(!fout)
    {
        perror(outfile.c_str());
        return -1;
    }
    
    if(audio.file.size())
    {
        if(!audio.spacing) audio.spacing = frameus * 2 / 1E6;
        if(audio.spacing < frameus / 1E6)
        {
            fprintf(stderr, "Error: Audio spacing (%g) can not be less than video frame length (%g).\n", audio.spacing, frameus/1E6);
            return -1;
        }
        if(audio.spacing > 1)
        {
            fprintf(stderr, "Warning: Audio spacing (%g) should usually be less than second.\n", audio.spacing);
        }
    }
    
    CalcSize();
    WriteAVI();
    
    fclose(fout);
    
    return 0;
}
